// EEPROM
#include <EEPROM.h>
#define EEPROM_SIZE 13 // 13 byte

// 크리스털 LCD
#include <Wire.h> 
#include <LiquidCrystal_I2C.h>
LiquidCrystal_I2C lcd(0x27, 16, 2); // D1 - SCL, D2 SDA, Set the LCD address to 0x27 for a 16 chars and 2 line display

// DFPlayer Software Serial
#include <SoftwareSerial.h> 
#define DF_rxPin 2 // DFplayer RX -> NodeMcu D4
#define DF_txPin 0 // DFplayer TX -> NodeMcu D3 
SoftwareSerial dfSerial(DF_txPin, DF_rxPin); // (RX, TX)

// wifi & NTP 라이브러리
#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <TimeLib.h>

// NTP 설정
#define ssid "SK_WiFiGIGA40F7" // your network SSID (name)
#define pass "1712037218"  // your network password
unsigned int localPort = 2390;      // local port to listen for UDP packets
char timeServer[] = "kr.pool.ntp.org";  // NTP server
const int NTP_PACKET_SIZE = 48; // NTP time stamp is in the first 48 bytes of the message
byte packetBuffer[ NTP_PACKET_SIZE];
WiFiUDP Udp;

// Soft AP 설정
#define AP_SSID  "Esp12"
#define AP_PASS  "1234test"
WiFiServer server(80);
WiFiClient client;

bool got_NTP = false;
uint8_t timeZone = 9;
bool wait_NTP = false;

// Soft AP용 변수
String income_wifi = ""; // 와이파이 수신 스트링 저장 변수
uint8_t received_df[10] = {0, }; // DFplayer에서 받은 데이터 저장
String df_value1 = "";  // 안드로이드에 전송할 값
String df_value2 = "";  // 안드로이드에 전송할 값

char buf[5]; // 스트링 헥사값 십진수 변경용 변수, FF -> 255

uint8_t df_count = 0;       // 연속된 값 받기위한 카운트
bool wait_receive = false;  // dfplayer에 질의후 대기시에만 dfSerial 확인용 플래그
bool convert = false;       // dfplayer의 데이터수신후 스트링 변환 진입 플래그

// DFplayer command 변수
uint8_t cmd_d; 
uint8_t ack_d;
uint8_t msb_d;
uint8_t lsb_d;

uint8_t df_command[] = {0x7E, 0xFF, 0x06, 0x00, 0x00, 0x00, 0x00, 0xEF};

void send_command(uint8_t cmd, uint8_t ack, uint8_t msb, uint8_t lsb) {
  df_command[3] = cmd;
  df_command[4] = ack;
  df_command[5] = msb;
  df_command[6] = lsb;
  for(int i = 0; i < sizeof(df_command); i++){
    dfSerial.write(df_command[i]);
  }
}

int h = 12; // initial Time display is 12:59:45 PM
int m = 56;
uint8_t s = 45;
bool meridian = true;
uint8_t hm;
bool now_am = true; //AM
uint16_t yy = 2019;
uint8_t mm = 1;
uint8_t dd = 1;
uint8_t week_num = 1;
String week_day = "SUN";
bool adjust = false;

bool display_t = false;
unsigned long int start_time = 0;
uint8_t volume = 10;

bool receive_status = false; // 질의 코드 0x42수신 여부 플래그
uint8_t status_count = 0;   // 질의값 저장 인덱스용 카운터
uint8_t status_val[7];         // 질의값 저장 배열
bool played = false;         // 질의결과 플래그 (재생중인지 아닌지)
bool status_check = false; // 질의 수신 완료 플래그

unsigned long int count_time = 0;
bool delay_set = false;
uint8_t sequence = 0;

bool speak = false; // 현재 상태 질의 시작 및 음성출력 시작 플래그
bool speak_finished = true;
bool short_sentence = false;
bool sharp = false;

int am;
int time_sharp;
int hour_a;
int min_w;
int min_s;

uint8_t alarm_track_folder = 2;
uint8_t alarm_track = 2;

bool alarm_on = false;
int alarm_h;
int alarm_m;
unsigned long int alarm_delay = 0;
bool alarm_sentence = false;
uint8_t alarm_count = 0;
bool alarm_time = false;

unsigned long int alarm_sequence_delay = 0;
uint8_t alarm_sequence = 0;
bool alarm_send_f = false;
bool alarm_stop = false; 

bool no_speak_sharp = false;
uint8_t no_speak_s = 0;
uint8_t no_speak_f = 0;

uint8_t time_volume = 30; // 음성출력 및 알람 볼륨

uint8_t pin = D5;

unsigned long int status_time = 0;
bool status_not_received = false;

bool snooze_on = false;
bool snooze_time = false;
uint8_t snooze_h = 0;
uint8_t snooze_m = 0;
uint8_t snooze_set = 5; // 스누즈 설정 시간

void setup() {
  Serial.begin(115200);
  dfSerial.begin(9600); // DFplayer Serial
  EEPROM.begin(EEPROM_SIZE);
  pinMode(pin, INPUT_PULLUP);  // pullup 저항 이용 - 버튼쪽에 저항 없음 OFF - HIHG, ON - LOW
  if(EEPROM.read(0) != 1) {
    EEPROM.write(0, 1); // eeprom 저장 플래그
    EEPROM.write(1, no_speak_sharp); // 묵음 플래그
    EEPROM.write(2, alarm_on); // 알람 플래그
    EEPROM.write(3, meridian); // 24 플래그
    EEPROM.write(4, no_speak_s); // 묵음 시작
    EEPROM.write(5, no_speak_f); // 묵음 종료
    EEPROM.write(6, alarm_h); // 알람 시
    EEPROM.write(7, alarm_m); // 알람 분
    EEPROM.write(8, alarm_track_folder); // 알람 폴더
    EEPROM.write(9, alarm_track); // 알람 트랙
    EEPROM.write(10, time_volume); // 알람 볼륨 time_volume
    EEPROM.write(11, snooze_on); // 스누즈 플래그
    EEPROM.write(12, snooze_set); // 스누즈 시간 
    EEPROM.commit();
  } else {
    no_speak_sharp = EEPROM.read(1);
    alarm_on = EEPROM.read(2);
    meridian = EEPROM.read(3);
    no_speak_s = EEPROM.read(4);
    no_speak_f = EEPROM.read(5);
    alarm_h = EEPROM.read(6);
    alarm_m = EEPROM.read(7);
    alarm_track_folder = EEPROM.read(8);
    alarm_track = EEPROM.read(9);
    time_volume = EEPROM.read(10);
    snooze_on = EEPROM.read(11);
    snooze_set = EEPROM.read(12);
  }
  WiFi.mode(WIFI_AP_STA);
  WiFi.begin(ssid, pass);
  lcd.begin();
  lcd.backlight();
  lcd.clear();
  lcd.setCursor(0,0);  // row, col의 좌표로 커서를 위치 col: 0 ~ 15 row: 0, 1
  lcd.print(F("Arduino clock"));
  lcd.setCursor(0,1);
  lcd.print(F("Connecting..."));
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(F("."));
  }
  Serial.println(F("You're connected to the network"));
  Udp.begin(localPort);
  WiFi.softAP(AP_SSID, AP_PASS);  // WiFi.softAP(ssid, password);
  IPAddress myIP = WiFi.softAPIP();
  Serial.print(F("AP IP address: "));
  Serial.println(myIP);
  server.begin();
  send_command(0x06, 0, 0, volume); // 기동시 초기 볼륨값 15
}

unsigned long int button_millis = 0;
bool button_hold = false;

void loop() {
  botton_delay();
  status_repeat();
  wifi_delay();
  NTP_send();
  getNtpTime();
  wait_NTP_time();
  cal_time();
  adjust_time();  // 시리얼 모니터 설정 함수
  speak_time();
  delayCount();
  alarm_warning();
  alarm_send();
  if (wait_receive == true || status_not_received == true) { 
    if (dfSerial.available()) {
      if (dfSerial.peek() == 0x7E) {  // 42 00 02 01
        uint8_t dump = dfSerial.readBytes(received_df, 10);
        if (speak_finished == false) {
          if (received_df[3] == 0x42) {
            status_check = true;
            if (received_df[5] == 2 && received_df[6] == 1) played = true; // 인덱스 2 = 2 -> SD카드, 인덱스 3 = 1 재생중
            else if (received_df[5] == 2 && received_df[6] == 0) played = false; // 인덱스 3 = 0 정지중
            status_not_received = false;
          }
        }
        else if (wait_receive == true) {
          df_count++;
          if (df_count == 2) wait_receive = false;
          convert = true;
        }
      }
      else {
        uint8_t dump = dfSerial.read(); // 코드 값 아닌경우 무시
      }
    }
  }
  if (convert == true) {
    if (df_count == 1) {
      for(int i = 0; i < 10; i++) {
        if (received_df[i] < 16) df_value1 += String('0') + String(received_df[i], HEX);
        else df_value1 += String(received_df[i], HEX);
      }
      received_df[10] = {0,};
    }
    else if (df_count == 2) {
      for(int i = 0; i < 10; i++) {
        if (received_df[i] < 16) df_value2 += String('0') + String(received_df[i], HEX);
        else df_value2 += String(received_df[i], HEX);
      }
      received_df[10] = {0,};
      df_count = 0;
    }
    convert = false;
  }
  wifi_delay(); // 와이파이 데이터 읽기
  if (button_hold == false) {  // button 한번만 읽도록 하는 조건
    if (digitalRead(pin) == LOW) { // LOW - ON
      button_hold = true;
      button_millis = millis();
      if (alarm_time == true) {    // 알람 재생중일 때 알람 정지
        alarm_stop = true;
        alarm_time = false;
        alarm_sentence = false;
        alarm_count = 0;
        alarm_send_f = true;
        if (wait_NTP == true) got_NTP = false;
        if (snooze_on) snooze();  // 스누즈 작동
      }
      else {  // 알람이 재생중이지 않을 때 현재 시간 출력
        speak_option();
      }
    }
  }
}

void botton_delay() {
  if (button_hold == true) {
    if (millis() - button_millis >= 500) button_hold = false;
  }
}

uint8_t button_status_count = 0;    
bool send_button_status = false;
uint8_t client_out_check = 0;
bool client_in = false;

unsigned long int one_millis = 0;

void wifi_delay() { // 30 밀리초 마다 client 연결 확인 및 데이터 수신
  if (millis() - one_millis > 30) {
    one_millis = millis();
    WiFiClient client;
    wifi_read();
    if (client_in == true) {
      if(!client.available()) {
        client_out_check++;
        if (client_out_check > 100) {
           button_status_count = 0;
           client_in = false;
           send_button_status = false;
           Serial.println("Client disconnected");
        }
      }
    }
  }
}

// 디지털 핀 제어용 변수
uint8_t pin_val = 0;

// 웹요청 응답용 변수
String send_pin = "";
String send_text = "";

void wifi_read() { // wifi 데이터 수신
  client = server.available(); // 서버가 작동하고 있으면 client 생성
  if(client.available()) {     // 클라이언트에 데이터가 있으면
    income_wifi = client.readStringUntil('\r');
    if (income_wifi.indexOf("7EFF06") != -1) {          // DFPlayer 시작 코드가 있으면
      String wifi_temp = "0x" + income_wifi.substring(income_wifi.indexOf("7EFF06")+6, income_wifi.indexOf("7EFF06")+8);
      wifi_temp.toCharArray(buf, 5);
      cmd_d = strtol(buf, NULL, 16);
      wifi_temp = "0x" + income_wifi.substring(income_wifi.indexOf("7EFF06")+8, income_wifi.indexOf("7EFF06")+10);
      wifi_temp.toCharArray(buf, 5);
      ack_d = strtol(buf, NULL, 16);
      wifi_temp = "0x" + income_wifi.substring(income_wifi.indexOf("7EFF06")+10, income_wifi.indexOf("7EFF06")+12);
      wifi_temp.toCharArray(buf, 5);
      msb_d = strtol(buf, NULL, 16);
      wifi_temp = "0x" + income_wifi.substring(income_wifi.indexOf("7EFF06")+12, income_wifi.indexOf("7EFF06")+14);
      wifi_temp.toCharArray(buf, 5);
      lsb_d = strtol(buf, NULL, 16);
      send_command(cmd_d, ack_d, msb_d, lsb_d);
      if (cmd_d == 0x06) { volume = lsb_d; Serial.println(volume); }
      wait_receive = true;
    }
    else if (income_wifi.indexOf("%%F1") != -1) {          // 종료문자가 있으면
      if (income_wifi.indexOf("%%F0") != -1) {        // 디지털 핀 값 수신
        String wifi_temp = income_wifi.substring(income_wifi.indexOf("%%F0")+4, income_wifi.indexOf("%%F1"));
        pin_val = wifi_temp.toInt();
        pin_control();
      }
    }
    else if (income_wifi.indexOf("%%F5") != -1) { // 스케쥴 접속확인 응답, 텍스트 및 로컬값 전송
      client.println(F("HTTP/1.1 200 OK\r\n"));      // 스케쥴 웹요청시 응답 헤더 
      client_out_check = 0;
      if (send_button_status == false) { 
        button_status_count++;
        if (button_status_count == 8) {
          clock_status(); 
          send_button_status = true; 
          client_in = true;
          Serial.print("sent button status");
        }
      }
      if (df_value1 != "") { 
        client.print(df_value1);  
        df_value1 = "";  
        df_count = 0;
      }
      if (df_value2 != "") { 
        client.print(df_value2);  
        df_value2 = "";  
        df_count = 0;
      }
      if (send_pin != "") {                       // 전송할 핀 상태 값이 있으면
        client.print(send_pin); 
        send_pin = "";
      }
      if (send_text != "") {                      // 전송할 텍스트가 있으면
        client.print(send_text);
        send_text = "";
      }
    }
    else if (income_wifi.indexOf("%%F9") != -1) { // 알람 세팅값 수신
      String wifi_temp = income_wifi.substring(income_wifi.indexOf("%%F9")+4, income_wifi.indexOf("\n"));
      String set = wifi_temp.substring(0, wifi_temp.indexOf(":"));
      alarm_h = set.toInt();
      set = wifi_temp.substring(wifi_temp.indexOf(":") + 1, wifi_temp.length());
      alarm_m = set.toInt();
      EEPROM.write(6, alarm_h); // 알람 시
      EEPROM.write(7, alarm_m); // 알람 분
      EEPROM.commit();
      client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
      client.print(F("alarm set ")); client.print(alarm_h); client.print(F(":")); client.print(alarm_m);
    }
    else {  // 텍스트 명령
      String wifi_temp = income_wifi.substring(income_wifi.indexOf("GET /")+5, income_wifi.indexOf("HTTP/1.1")-1);
      if (wifi_temp == "dtime") display_time_serial(); //
      else if (wifi_temp == "s") { s = 0; start_time = millis(); display_t = true; }
      else if (wifi_temp == "mm") { m++; display_t = true; }
      else if (wifi_temp == "m") { m--; display_t = true; }
      else if (wifi_temp == "hh") { h++; display_t = true; }
      else if (wifi_temp == "h") { h--; display_t = true; }
      else if (wifi_temp == "24") { meridian = !meridian; EEPROM.write(3, meridian); EEPROM.commit(); display_t = true; }
      else if (wifi_temp == "yy") { yy++; display_t = true; }
      else if (wifi_temp == "y") { yy--; display_t = true; }
      else if (wifi_temp == "mon") { mm++; adjust = true; display_t = true; }
      else if (wifi_temp == "mon-") { mm--; adjust = true; display_t = true; }
      else if (wifi_temp == "dd") { dd++; adjust = true; display_t = true; }
      else if (wifi_temp == "d") { dd--; adjust = true; display_t = true; }
      else if (wifi_temp == "ww") { week_num++; adjust = true; display_t = true; }
      else if (wifi_temp == "w") { week_num--; adjust = true; display_t = true; }
      else if (wifi_temp == "ntp") got_NTP = false;
      else if (wifi_temp == "nsp") { 
        if (no_speak_s == no_speak_f) {
          client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
          client.println(F("Please set NSP time"));
        }
        else {
          no_speak_sharp = !no_speak_sharp; 
          EEPROM.write(1, no_speak_sharp); 
          EEPROM.commit();
          client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
          if (no_speak_sharp) client.print(F("NSP ON "));
          else client.print(F("NSP OFF "));
          client.print(F("no speak time: ")); client.print(no_speak_s); client.print("-"); client.println(no_speak_f); 
        }
      }
      else if (wifi_temp.startsWith("atrack")) {
        String set = wifi_temp.substring(6, wifi_temp.indexOf(","));
        alarm_track_folder = set.toInt();
        set = wifi_temp.substring(wifi_temp.indexOf(",") + 1, wifi_temp.length());
        alarm_track = set.toInt();
        EEPROM.write(8, alarm_track_folder); // 알람 폴더
        EEPROM.write(9, alarm_track); // 알람 트랙
        EEPROM.commit();
        client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
        client.print(F("Folder: ")); client.print(alarm_track_folder); client.print(" ");
        client.print(F("Track: ")); client.print(alarm_track); 
      }
      else if (wifi_temp.startsWith("avol")) {
        wifi_temp.remove(0, 4);
        time_volume = wifi_temp.toInt();
        EEPROM.write(10, time_volume); // 알람 볼륨
        EEPROM.commit();
        client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
        client.print(F("alarm volume: ")); client.print(time_volume); 
      }
      else if (wifi_temp.startsWith("tset")) {
        String set = wifi_temp.substring(4, 6);
        h = set.toInt();
        if (h > 23) h = 0;
        set = wifi_temp.substring(6, 8);
        m = set.toInt();
        if (m > 59) m = 0;
        set = wifi_temp.substring(8, 10);
        s = set.toInt();
        if (s > 59) s = 0;
        start_time = millis();
        adjust = true;
        display_t = true;
      }
      else if (wifi_temp.startsWith("dset")) {
        String set = wifi_temp.substring(4, 8);
        yy = set.toInt();
        set = wifi_temp.substring(8, 10);
        mm = set.toInt();
        if (mm > 12) mm = 1;
        set = wifi_temp.substring(10, 12);
        dd = set.toInt();
        if (dd > 31) dd = 0;
        adjust = true;
        display_t = true;
      }
      else if (wifi_temp == "alarm") { 
        alarm_on = !alarm_on;
        EEPROM.write(2, alarm_on); // 알람 플래그
        EEPROM.commit();
        client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
        if (alarm_on) client.print(F("alarm on"));
        else client.print(F("alarm off"));
      }
      else if (wifi_temp == "snooze") { 
        snooze_on = !snooze_on;
        EEPROM.write(11, snooze_on); // 알람 플래그
        EEPROM.commit();
        client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
        if (alarm_on) client.print(F("Snooze ON"));
        else client.print(F("Snooze OFF"));
      }
      else if (wifi_temp.startsWith("zset")) { // 스누즈 시간 설정
        String set = wifi_temp.substring(4, 6);
        snooze_set = set.toInt();
        if (snooze_set > 59) snooze_set = 59; 
        EEPROM.write(12, snooze_set); // 알람 플래그
        EEPROM.commit();
        client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
        client.print(F("snooze_set Time: ")); client.print(snooze_set);
      }
      else if (wifi_temp.startsWith("aset")) {
        String set = wifi_temp.substring(4, 6);
        alarm_h = set.toInt();
        if (alarm_h > 23) alarm_h = 0;
        set = wifi_temp.substring(6, 8);
        alarm_m = set.toInt();
        if (alarm_m > 59) alarm_m = 0;
        EEPROM.write(6, alarm_h); // 알람 시
        EEPROM.write(7, alarm_m); // 알람 분
        EEPROM.commit();
        client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
        client.print(F("alarm set ")); client.print(alarm_h); client.print(F(":")); client.print(alarm_m); 
      }
      else if (wifi_temp.startsWith("setnsp")) {
        String set = wifi_temp.substring(6, wifi_temp.indexOf("~"));
        no_speak_s = set.toInt();
        if (no_speak_s > 23) no_speak_s = 0;
        set = wifi_temp.substring(wifi_temp.indexOf("~") + 1, wifi_temp.length());
        no_speak_f = set.toInt();
        if (no_speak_f > 23) no_speak_f = 0;
        EEPROM.write(4, no_speak_s); // 묵음 시작
        EEPROM.write(5, no_speak_f); // 묵음 종료
        EEPROM.commit();
        client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
        client.print(F("no speak set: ")); client.print(no_speak_s); client.print("-"); client.print(no_speak_f); 
      }
      else if (wifi_temp == "test") speak_option();
      else if (wifi_temp == "stop") { 
        alarm_stop = true;
        alarm_time = false;
        alarm_sentence = false;
        alarm_count = 0;
        alarm_send_f = true;
        if (wait_NTP == true) got_NTP = false;
        snooze_time = false;
      }
      else if (wifi_temp == "play") send_command(0x0F, 0, 2, 6);
      else if (wifi_temp == "pause") send_command(0x0E, 1, 0, 0);
      else if (wifi_temp == "previous") send_command(0x02, 1, 0, 0);
      else if (wifi_temp == "next") send_command(0x01, 1, 0, 0);
      else if (wifi_temp == "vv") { if(volume < 30) volume++; send_command(0x06, 0, 0, volume); }
      else if (wifi_temp == "v") { if(volume > 0) volume--; send_command(0x06, 0, 0, volume); }
      else if (wifi_temp == "reset") send_command(0x0C, 0, 0, 0);
      if (m >= 60) m = 0;
      else if (m < 0) m = 59;
      if (h >= 24) h = 0; 
      else if (h < 0) h = 23; 
      if (h < 12) now_am = true;
      else now_am = false; 
      hm = h; 
      if(meridian == true && hm == 12) { now_am = false; }
      else if (meridian == true && hm > 12) { hm = hm - 12; }
      cal_dd();
      if (display_t == true) { display_time_serial(); display_t = false; } // display_time();
    }
  }
  income_wifi = "";
}

void adjust_time() {
  if (Serial.available() > 0) {
    String temp = Serial.readStringUntil('\n');
    if (temp == "dtime") display_time();
    else if (temp == "s") { s = 0; start_time = millis(); display_t = true; }
    else if (temp == "mm") { m++; display_t = true; }
    else if (temp == "m") { m--; display_t = true; }
    else if (temp == "hh") { h++; display_t = true; }
    else if (temp == "h") { h--; display_t = true; }
    else if (temp == "24") { meridian = !meridian; EEPROM.write(3, meridian); EEPROM.commit(); display_t = true; }
    else if (temp == "yy") { yy++; display_t = true; }
    else if (temp == "y") { yy--; display_t = true; }
    else if (temp == "mon") { mm++; adjust = true; display_t = true; }
    else if (temp == "mon-") { mm--; adjust = true; display_t = true; }
    else if (temp == "dd") { dd++; adjust = true; display_t = true; }
    else if (temp == "d") { dd--; adjust = true; display_t = true; }
    else if (temp == "ww") { week_num++; adjust = true; display_t = true; }
    else if (temp == "w") { week_num--; adjust = true; display_t = true; }
    else if (temp == "ntp") got_NTP = false;
    else if (temp == "nsp") { 
      if (no_speak_s == no_speak_f) Serial.println(F("Please set NSP time"));
      else {
        no_speak_sharp = !no_speak_sharp; 
        EEPROM.write(1, no_speak_sharp);
        EEPROM.commit();
        if (no_speak_sharp) Serial.println(F("NSP ON"));
        else Serial.println(F("NSP OFF"));
        Serial.print(F("no speak time: ")); Serial.print(no_speak_s); Serial.print("-"); Serial.println(no_speak_f); 
      }
    }
    else if (temp.startsWith("atrack")) {
      String set = temp.substring(6, temp.indexOf(","));
      alarm_track_folder = set.toInt();
      set = temp.substring(temp.indexOf(",") + 1, temp.length());
      alarm_track = set.toInt();
      EEPROM.write(8, alarm_track_folder); // 알람 폴더
      EEPROM.write(9, alarm_track); // 알람 트랙
      EEPROM.commit();
      Serial.print(F("Folder: ")); Serial.print(alarm_track_folder); Serial.print(" ");
      Serial.print(F("Track: ")); Serial.println(alarm_track); 
    }
    else if (temp.startsWith("avol")) {
      temp.remove(0, 4);
      time_volume = temp.toInt();
      EEPROM.write(10, time_volume); // 알람 볼륨
      EEPROM.commit();
      Serial.print(F("alarm volume: ")); Serial.println(time_volume); 
    }
    else if (temp.startsWith("tset")) {
      String set = temp.substring(4, 6);
      h = set.toInt();
      if (h > 23) h = 0;
      set = temp.substring(6, 8);
      m = set.toInt();
      if (m > 59) m = 0;
      set = temp.substring(8, 10);
      s = set.toInt();
      if (s > 59) s = 0;
      start_time = millis();
      adjust = true;
      display_t = true;
    }
    else if (temp.startsWith("dset")) {
      String set = temp.substring(4, 8);
      yy = set.toInt();
      set = temp.substring(8, 10);
      mm = set.toInt();
      if (mm > 12) mm = 1;
      set = temp.substring(10, 12);
      dd = set.toInt();
      if (dd > 31) dd = 0;
      adjust = true;
      display_t = true;
    }
    else if (temp == "alarm") { 
      alarm_on = !alarm_on;
      EEPROM.write(2, alarm_on); // 알람 플래그
      EEPROM.commit();
      if (alarm_on) Serial.println(F("alarm on"));
      else Serial.println(F("alarm off"));
    }
    else if (temp == "snooze") { 
      snooze_on = !snooze_on;
      EEPROM.write(11, snooze_on); // 알람 플래그
      EEPROM.commit();
      if (snooze_on) Serial.println(F("Snooze ON"));
      else Serial.println(F("Snooze OFF"));
    }
    else if (temp.startsWith("zset")) { // 스누즈 시간 설정
      String set = temp.substring(4, 6);
      snooze_set = set.toInt();
      if (snooze_set > 59) snooze_set = 59; 
      EEPROM.write(12, snooze_set); // 알람 플래그
      EEPROM.commit();
      Serial.print(F("snooze_set Time: ")); Serial.println(snooze_set);
    }
    else if (temp.startsWith("aset")) {
      String set = temp.substring(4, 6);
      alarm_h = set.toInt();
      if (alarm_h > 23) alarm_h = 0;
      set = temp.substring(6, 8);
      alarm_m = set.toInt();
      if (alarm_m > 59) alarm_m = 0;
      EEPROM.write(6, alarm_h); // 알람 시
      EEPROM.write(7, alarm_m); // 알람 분
      EEPROM.commit();
      Serial.print("alarm set "); Serial.print(alarm_h); Serial.print(":"); Serial.println(alarm_m); 
    }
    else if (temp.startsWith("setnsp")) {
      String set = temp.substring(6, temp.indexOf("~"));
      no_speak_s = set.toInt();
      if (no_speak_s > 23) no_speak_s = 0;
      set = temp.substring(temp.indexOf("~") + 1, temp.length());
      no_speak_f = set.toInt();
      if (no_speak_f > 23) no_speak_f = 0;
      EEPROM.write(4, no_speak_s); // 묵음 시작
      EEPROM.write(5, no_speak_f); // 묵음 종료
      EEPROM.commit();
      Serial.print(F("no speak set: ")); Serial.print(no_speak_s); Serial.print("-"); Serial.println(no_speak_f); 
    }
    else if (temp == "test") speak_option();
    else if (temp == "stop") { 
      alarm_stop = true;
      alarm_time = false;
      alarm_sentence = false;
      alarm_count = 0;
      alarm_send_f = true;
      if (wait_NTP == true) got_NTP = false;
      snooze_time = false;
    }
    else if (temp == "play") send_command(0x0F, 0, 2, 6);
    else if (temp == "pause") send_command(0x0E, 1, 0, 0);
    else if (temp == "previous") send_command(0x02, 1, 0, 0);
    else if (temp == "next") send_command(0x01, 1, 0, 0);
    else if (temp == "vv") { if(volume < 30) volume++; send_command(0x06, 0, 0, volume); }
    else if (temp == "v") { if(volume > 0) volume--; send_command(0x06, 0, 0, volume); }
    else if (temp == "reset") send_command(0x0C, 0, 0, 0);
    if (m >= 60) m = 0;
    else if (m < 0) m = 59;
    if (h >= 24) h = 0; 
    else if (h < 0) h = 23; 
    if (h < 12) now_am = true;
    else now_am = false; 
    hm = h; 
    if(meridian == true && hm == 12) { now_am = false; }
    else if (meridian == true && hm > 12) { hm = hm - 12; }
    cal_dd();
    if (display_t == true) { display_time_serial(); display_t = false; } // 시리얼 모니터 출력
  }
}

void clock_status() {  // 와이파이 연결시 버튼 및 볼륨 상태값 전송
  String volume_s = "";
  if (volume < 16) volume_s = '0' + String(volume, HEX);
  else volume_s = String(volume, HEX);
  df_value1 = "7eff06430000" + volume_s + "0000ef"; // 현재 볼륨
  if (no_speak_sharp == false) send_pin_echo(10);
  else send_pin_echo(11);
  if (alarm_on == false) send_pin_echo(20);
  else send_pin_echo(21);
  if (meridian == true) send_pin_echo(30);
  else send_pin_echo(31);
  if (snooze_on == false) send_pin_echo(80);
  else send_pin_echo(81);
}

void speak_option() {
  speak_failed();
  if (m == 0 && s != 0) {
    sharp = true;
    if (h == 0) time_sharp = 2212;
    else if (h < 13 && h > 0) time_sharp = 2200 + h;
    else time_sharp = 2200 + (h - 12);
  }
  else {
    sharp = false;
    if (h == 0) hour_a = 1012;
    else if (h < 13 && h > 0) hour_a = 1000 + h;
    else hour_a = 1000 + (h - 12);
    if (h < 12) am = 1301;
    else am = 1302;
    int temp = m%10;
    if (temp == 0 || m < 10) {
      short_sentence = true;
      min_s = 1200 + m;
    }
    else {
      short_sentence = false;
      min_w = 1100 + (m - temp);
      min_s = 1200 + temp;
    }
  }
  speak = true;
  speak_finished = false;
}

uint8_t am_s = 4;
uint8_t vol_s = 6;
uint8_t hour_a_s = 21;
uint8_t long_min_w_s = 44;
uint8_t long_min_s_s = 65;
uint8_t long_min_s_end = 96;
uint8_t short_min_s_s = 44;
uint8_t short_min_s_end = 74;
uint8_t time_sharp_s = 4;
uint8_t time_sharp_end = 80;

void status_repeat() {
  if ( status_not_received == true) {
    if (millis() - status_time >= 50) {
      status_time = millis();
      send_command(0x42, 0, 0, 0);
      Serial.print(".");
    }
  }
}

void speak_time() {
  if (speak == true) {
    if (sequence == 0 && status_check == false) {
      Serial.println("speak time");
      status_not_received = true;
      send_command(0x42, 0, 0, 0);
      sequence++;
      status_time = millis();
    }
    else if (sequence == 1 && status_check == true) { // status_not_received = false; -> status_check = true;
      if (played == false) {
        send_command(0x0F, 0, 2, 255);
        sequence++; 
        delay_set = true;
      }
      else {
        delay_set = true;
        sequence = 3;
      }
    }
    else if (sequence >= am_s && sharp == false) {
      if (sequence == am_s) { advert(am); speak = false; }
      else if (sequence == vol_s) { send_command(0x06, 0, 0, time_volume); speak = false; }
      else if (sequence == hour_a_s) { advert(hour_a); speak = false; }
      if (short_sentence == false) {
        if (sequence == long_min_w_s) { advert(min_w); speak = false; }
        else if (sequence == long_min_s_s) { advert(min_s); speak = false; }
        else if (sequence == long_min_s_end) {
          send_command(0x06, 0, 0, volume);
          speak_finished = true;
          speak = false;
          status_check = false;
          sequence = 0;
          if (wait_NTP == true) got_NTP = false;
        }
      }
      else {
        if (sequence == short_min_s_s) { advert(min_s); speak = false; }
        else if (sequence == vol_s) { send_command(0x06, 0, 0, time_volume); speak = false; }
        else if (sequence == short_min_s_end) {
          send_command(0x06, 0, 0, volume);
          speak_finished = true;
          speak = false;
          status_check = false;
          sequence = 0;
          if (wait_NTP == true) got_NTP = false;
        }
      }
    }
    else if (sequence >= time_sharp_s && sharp == true) {
      if (sequence == time_sharp_s) { advert(time_sharp); speak = false; }
      else if (sequence == vol_s) { send_command(0x06, 0, 0, time_volume); speak = false; }
      else if (sequence == time_sharp_end) {
        send_command(0x06, 0, 0, volume);
        speak_finished = true;
        speak = false;
        status_check = false;
        sequence = 0;
        time_sharp_end = 80;
        if (wait_NTP == true) got_NTP = false;
      }
    }
  }
}

void advert(int ad_num) {
  uint8_t ad_lsb = ad_num;
  uint8_t ad_msb = ad_num >> 8;
  send_command(0x13, 0, ad_msb, ad_lsb);
}

void delayCount() {
  if (delay_set == true) {
    if (millis() - count_time >= 50) { // 시간 간격: 밀리초
      count_time = millis();
      sequence++;
      if (sharp == false) { 
        if (sequence == am_s) speak = true;
        else if (sequence == vol_s) speak = true;
        else if (sequence == hour_a_s) speak = true;
        if (short_sentence == false) {
          if (sequence == long_min_w_s) speak = true;
          else if (sequence == long_min_s_s) speak = true;
          else if (sequence == long_min_s_end) { speak = true; delay_set = false; }
        }
        else {
          if (sequence == short_min_s_s) speak = true;
          else if (sequence == short_min_s_end) { speak = true; delay_set = false; }
        }
      }
      else {
        if (sequence == time_sharp_s) speak = true;
        else if (sequence == vol_s) speak = true;
        else if (sequence == time_sharp_end)  { speak = true; delay_set = false; }
      }
    }
  }
}

uint8_t change_millis = 0; 
uint16_t millis_sec = 1000;

void cal_time() {
  if (millis() - start_time >= millis_sec) { // 시간 간격: 밀리초
    start_time = millis(); // 상기 조건을 만족할때의 밀리초를 다시 start_time에 저장하여 조건 초기화
    change_millis++;
    if (change_millis < 60) millis_sec = 1000;
    else { millis_sec = 1001; change_millis = 0; }
    s++;
    if (s == 60) {
      s = 0;
      m++;
    }
    if(m == 60) { m = 0; h++; }
    if(h == 12) { now_am = false;}
    else if (h == 24) { h = 0; now_am = true; dd++; week_num++; cal_dd(); }
    hm = h;
    if(meridian == true && hm >= 12) { now_am = false; }
    else now_am = true;
    if (meridian == true && hm > 12) { hm = hm - 12; }
    display_time();          // 크리스털 LCD 출력함수
    display_time_serial();   // 시리얼 모니터 출력함수 - 주석처리시 display_t 옵션에 의해 출력
    if (alarm_on) {
      if (alarm_h == h && alarm_m == m && s == 0){
        alarm_time = true;
        alarm_send_f = true;
      }
      if (snooze_time) {
        if (snooze_h == h && snooze_m == m && s == 0){
          alarm_time = true;
          alarm_send_f = true;
        }
      }
    }
    if (alarm_time != true) {
      if (m == 0 && s == 0) {
        if (no_speak_sharp == true) {
          if (no_speak_s == no_speak_f) {
            no_speak_sharp = false;
            speak_sharp();
          }
          else {
            if (no_speak_s > no_speak_f) {
              if (h < no_speak_s && h > no_speak_f) {
                speak_sharp();
              }
            }
            else if (no_speak_s < no_speak_f) {
              if (h < no_speak_s && h >= 0 || h > no_speak_f && h <= 24) {
                speak_sharp();
              }
            }
          }
        }
        else {
          speak_sharp();
        }
      }
    }
    if (h == 1 && m == 1 && s == 10) {
      if (alarm_time == true || speak == true) wait_NTP = true;
      else {
        got_NTP = false; 
      }
    }
  }
}

void snooze() {
  snooze_time = true;
  snooze_m = m; // 버튼 정지시 현재 시간
  snooze_h = h;
  snooze_m = snooze_m + snooze_set;
  if (snooze_m > 59) {
    snooze_h = snooze_h + 1;
    snooze_m = snooze_m - 60;
    if (snooze_h > 23) snooze_h = 0;
  }
}

void wait_NTP_time() {
  if (wait_NTP == true) {
    got_NTP = true; 
    wait_NTP = false;
  }
}

void speak_failed() {
  if (sequence == 1 && status_check == false) { 
    sequence = 0; 
    receive_status = false; 
    speak = false;
    speak_finished = true;
  }
}

void speak_sharp() {
  speak_failed();
  if (speak_finished == true) {
    sharp = true;
    if (h == 0) time_sharp = 2112;
    else if (h < 13 && h > 0) time_sharp = 2100 + h;
    else time_sharp = 2100 + (h - 12);
    time_sharp_end = 48;
    speak = true;   
    speak_finished = false;
  }
}

void alarm_send() {
  if (alarm_send_f == true) {
    if (millis() - alarm_sequence_delay >= 50) { // 시간 간격: 밀리초
      alarm_sequence_delay = millis();
      alarm_sequence++;
      if (alarm_stop == true) {
        if (alarm_sequence == 1) send_command(0x06, 0, 0, volume);
        else if (alarm_sequence == 2) send_command(0x16, 0, 0, 0); 
        else if (alarm_sequence == 3) {
          send_command(0x15, 0, 0, 0); 
          alarm_send_f = false;
          alarm_sequence = 0;
          alarm_stop = false;
        }
      }
      else {
        if (alarm_sequence == 1) send_command(0x16, 0, 0, 0);
        else if (alarm_sequence == 2) send_command(0x06, 0, 0, time_volume); // 볼륨 30
        else if (alarm_sequence == 3) send_command(0x0F, 0, alarm_track_folder, alarm_track); // 폴더2 002번 사운드 파일
        else if (alarm_sequence == 8) {
          send_command(0x19, 0, 0, 0); // 반복재생
          alarm_sentence = true;
          alarm_count = 0;
          alarm_send_f = false;
          alarm_sequence = 0;
        }
      }
    }
  }
}

void alarm_warning() {
  if (alarm_sentence == true) {
    if (millis() - alarm_delay >= 1000) { // 시간 간격: 밀리초
      alarm_delay = millis();
      alarm_count++;
      if (alarm_count == 4) {
        advert(1401); 
      }
      else if (alarm_count == 12) {
        advert(1402);
      }
      else if (alarm_count == 25) {
        advert(1403);
      }
      else if (alarm_count == 70) {
        advert(1404);
        alarm_sentence = false;
        alarm_count = 0;
      }
    }
  }
}

void cal_dd() {
  uint8_t m_dds;
  switch (mm) {
    case 1:   m_dds = 31; break;
    case 2:   m_dds = 28; break;
    case 3:   m_dds = 31; break;
    case 4:   m_dds = 30; break;
    case 5:   m_dds = 31; break;
    case 6:   m_dds = 30; break;
    case 7:   m_dds = 31; break;
    case 8:   m_dds = 31; break;
    case 9:   m_dds = 30; break;
    case 10:  m_dds = 31; break;
    case 11:  m_dds = 30; break;
    case 12:  m_dds = 31; break;
  }
  if (adjust == false) {
    if (dd > m_dds) { dd = 1; mm++; }
    if (mm > 12) { mm = 1; yy++; }
    if (week_num > 7) week_num = 1;
    week_day_converter();
  }
  else {
    if (dd > m_dds) dd = 1;
    else if (dd < 1) dd = m_dds;
    if (mm > 12) mm = 1;
    else if (mm < 1) mm = 12;
    if (week_num > 7) week_num = 1;
    else if (week_num < 1) week_num = 7; 
    week_day_converter();
    adjust = false;
  }
}

void week_day_converter() {
  switch (week_num) {
    case 1:   week_day = "SUN"; break;
    case 2:   week_day = "MON"; break;
    case 3:   week_day = "TUE"; break;
    case 4:   week_day = "WED"; break;
    case 5:   week_day = "THU"; break;
    case 6:   week_day = "FRI"; break;
    case 7:   week_day = "SAT"; break;
  }
}

void display_time() {
  lcd.setCursor(0,0); 
  lcd.print(yy); lcd.print("."); 
  if (mm < 10) { lcd.print("0"); lcd.print(mm); lcd.print("."); }
  else { lcd.print(mm); lcd.print("."); }
  if (dd < 10) { lcd.print("0"); lcd.print(dd); lcd.print(". "); }
  else { lcd.print(dd); lcd.print(". "); }
  lcd.print(week_day); lcd.print(" ");  
  lcd.setCursor(0,1);
  if (meridian == true) {
    if (now_am == true) lcd.print("AM ");
    else lcd.print("PM ");
    if (hm < 10) { lcd.print("0"); lcd.print(hm); lcd.print(":"); }
    else { lcd.print(hm); lcd.print(":"); }
  }
  else {
    if (h < 10) { lcd.print("0"); lcd.print(h); lcd.print(":"); }
    else { lcd.print(h); lcd.print(":"); }
  }
  if (m < 10) { lcd.print("0"); lcd.print(m); lcd.print(":"); }
  else { lcd.print(m); lcd.print(":"); }
  if (s < 10) { lcd.print("0"); lcd.print(s); lcd.print("     "); }
  else { lcd.print(s); lcd.print("     "); }  // 크리스탈 LCD에는 pritln 명령어 없음
}

void display_time_serial() {
  Serial.print(yy); Serial.print("."); 
  if (mm < 10) { Serial.print("0"); Serial.print(mm); Serial.print("."); }
  else { Serial.print(mm); Serial.print("."); }
  if (dd < 10) { Serial.print("0"); Serial.print(dd); Serial.print(". "); }
  else { Serial.print(dd); Serial.print(". "); }
  Serial.print(week_day); Serial.print(" ");  
  if (meridian == true) {
    if (now_am == true) Serial.print("AM ");
    else Serial.print("PM ");
    if (hm < 10) { Serial.print("0"); Serial.print(hm); Serial.print(":"); }
    else { Serial.print(hm); Serial.print(":"); }
  }
  else {
    if (h < 10) { Serial.print("0"); Serial.print(h); Serial.print(":"); }
    else { Serial.print(h); Serial.print(":"); }
  }
  if (m < 10) { Serial.print("0"); Serial.print(m); Serial.print(":"); }
  else { Serial.print(m); Serial.print(":"); }
  if (s < 10) { Serial.print("0"); Serial.println(s); }
  else Serial.println(s);
}

void pin_control() {
  if (pin_val == 11 && no_speak_s == no_speak_f) {
    pin_val = 0;
    client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
    send_pin_echo(10);
    client.print(F("Set NSP ")); client.print(no_speak_s); client.print(F("-")); client.print(no_speak_f);
  }
  if (pin_val != 0) {
    switch (pin_val) {
      case 11: no_speak_sharp = true; 
               EEPROM.write(1, no_speak_sharp); // 묵음 플래그
               EEPROM.commit();
               client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
               client.print(F("NSP ON "));
               client.print(no_speak_s);
               client.print(F("-"));
               client.print(no_speak_f);
                  break;
      case 10: no_speak_sharp = false; 
               EEPROM.write(1, no_speak_sharp); // 묵음 플래그
               EEPROM.commit();
               client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
               client.print(F("NSP OFF "));
               client.print(no_speak_s);
               client.print(F("-"));
               client.print(no_speak_f);
                  break;
      case 21: alarm_on = true;
               EEPROM.write(2, alarm_on); // 알람 플래그
               EEPROM.commit();
               client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
               client.print(F("Alarm ON "));
               client.print(alarm_h);
               client.print(F(":"));
               client.print(alarm_m);
                  break;
      case 20: alarm_on = false;
               EEPROM.write(2, alarm_on); // 알람 플래그
               EEPROM.commit();
               client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
               client.print(F("Alarm OFF "));
               client.print(alarm_h);
               client.print(F(":"));
               client.print(alarm_m);
                  break;
      case 31: meridian = false;
               EEPROM.write(3, meridian); // 24 플래그
               EEPROM.commit();
                  break;
      case 30: meridian = true;
               EEPROM.write(3, meridian); // 24 플래그
               EEPROM.commit();
                  break;
      case 41: client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
               client.print(F("Folder: "));              
               client.print(alarm_track_folder);
               client.print(" ");
               client.print(F("Track: "));
               client.print(alarm_track);
                  break;
      case 51: alarm_stop = true; 
               alarm_time = false;
               alarm_sentence = false;
               alarm_count = 0;
               alarm_send_f = true;
               if (wait_NTP == true) got_NTP = false;
               snooze_time = false;
                  break;
      case 61: client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
               client.print(F("alarm volume: "));
               client.print(time_volume);
                  break;
      case 71: speak_option();
                  break;
      case 81: snooze_on = true;
               EEPROM.write(11, snooze_on); // 스누즈 플래그
               EEPROM.commit();
               client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
               client.print(F("Snooze ON"));
                  break;
      case 80: snooze_on = false;
               EEPROM.write(11, snooze_on); // 알람 플래그
               EEPROM.commit();
               client.println(F("HTTP/1.1 200 OK\r\n")); // 사용자 웹요청시 응답 헤더
               client.print(F("Snooze OFF"));
                  break;
    }
  pin_val = 0;
  }
}

void send_pin_echo(uint8_t pin_val){  // 디지털 버튼 상태값 전송
  String s_temp ="";
  s_temp += "%%F0";
  s_temp += pin_val;
  s_temp += "%%P1";
  if (client.connected())  client.print(s_temp); // 사용자 웹요청 응답
  else  send_pin = s_temp;                       // 스케쥴 웹요청 응답
}

unsigned long startMs = 0;
uint8_t kill_ntp = 0;

void NTP_send() {
  if (got_NTP == false) { // NTP 수신값이 없으면 2초마다 요청
    if (millis() - startMs >=  2000) {
      startMs = millis();
      Udp.begin(localPort);
      kill_ntp++;
      sendNTPpacket(timeServer);
      if (kill_ntp == 10) { got_NTP = false; kill_ntp = 0; }
    }
  }
}

void getNtpTime() {
  if (got_NTP == false) {
    if (Udp.available()) {
      if (Udp.parsePacket()) {
        Udp.read(packetBuffer, NTP_PACKET_SIZE);
        unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
        unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
        unsigned long secsSince1900 = highWord << 16 | lowWord;
        const unsigned long seventyYears = 2208988800UL;
        unsigned long epoch = secsSince1900 - seventyYears + timeZone * 3600;
        yy = year(epoch);
        mm = month(epoch);
        dd = day(epoch); 
        h = hour(epoch); 
        m = minute(epoch); 
        s = second(epoch); 
        week_num = weekday(epoch);
        week_day_converter();
        got_NTP = true;
        wait_NTP = false;
      }
    }
  }
}

void sendNTPpacket(char *ntpSrv) {
  memset(packetBuffer, 0, NTP_PACKET_SIZE);
  packetBuffer[0] = 0b11100011;   // LI, Version, Mode
  packetBuffer[1] = 0;     // Stratum, or type of clock
  packetBuffer[2] = 6;     // Polling Interval
  packetBuffer[3] = 0xEC;  // Peer Clock Precision
  packetBuffer[12]  = 49;
  packetBuffer[13]  = 0x4E;
  packetBuffer[14]  = 49;
  packetBuffer[15]  = 52;
  Udp.beginPacket(ntpSrv, 123); //NTP requests are to port 123
  Udp.write(packetBuffer, NTP_PACKET_SIZE);
  Udp.endPacket();
}
